@@ -11,7 +11,7 @@ my $cv = AE::cv();
 my $redis = AnyEvent::Redis::RipeRedis->new(
   host     => 'localhost',
   port     => '6379',
-  password => 'yourpass',
+  password => 'redis_pass',
 
   on_connect => sub {
     print "Connected to Redis server\n";
@@ -132,6 +132,14 @@ $redis->del( qw( foo bar list ),
 # Disconnect
 $redis->quit(
   sub {
+    shift;
+
+    if ( @_ ) {
+      my $err_msg = shift;
+
+      warn "$err_msg\n";
+    }
+
     $cv->send();
   }
 );
